/*
    pcb-rnd-3dview - watch a file for changes and run a viewer bg-process
    Copyright (C) 2021  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://www.repo.hu/projects/pcb-rnd-3dview
    VCS: svn://svn.repo.hu/pcb-rnd-3dview/trunk
    Contact: http://igor2.repo.hu/contact.html
*/

#include <stdio.h>
#include <string.h>
#include <poll.h>
#include <time.h>
#include <sys/time.h>
#include <sys/stat.h>

#include "projview.h"

projview_t prj;
const char *ifn, *icmd;
struct timespec ichg;

#define PRJ_TICK_S 0.05
#define FILE_TICK_S 0.5

double dtime(void)
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return (double)tv.tv_usec / 1000000.0 + (double)tv.tv_sec;
}

static void check_file(void)
{
	struct stat st;
	if (stat(ifn, &st) != 0) return;

	if (memcmp(&ichg, &st.st_mtim, sizeof(ichg)) != 0) {
		if (icmd != NULL)
			projview_load_proc(&prj, icmd);
		else
			projview_load_file(&prj, ifn);
		projview_cam(&prj);
		memcpy(&ichg, &st.st_mtim, sizeof(ichg));
	}
}

static void main_loop(void)
{
	struct pollfd pfds[1];
	double prj_next_tick = 0, file_next_tick = 0, now;


	pfds[0].fd = prj.prj_rfd;
	pfds[0].events = POLLIN;
	for(;;) {
		poll(pfds, sizeof(pfds)/sizeof(pfds[0]), 25);
		if (pfds[0].revents & POLLHUP)
			return;
		if (pfds[0].revents & POLLIN)
			projview_read(&prj);

		now = dtime();
		if (now >= prj_next_tick) {
			prj_next_tick = now + PRJ_TICK_S;
			projview_tick(&prj);
		}
		if (now >= file_next_tick) {
			file_next_tick = now + FILE_TICK_S;
			check_file();
		}
	}
}

static int help(const char *prg, const char *topic)
{
	printf("pcb-rnd-3dview - manage an external 3d viewer window\n");
	printf("\nUsage:\n");
	printf("%s -f file [-c cmd]\n", prg);
	printf("\n");
	printf("  -f     Watch file for changes; without -c: reload scene from file\n");
	printf("  -c     If specified, load scene by reading stdout of cmd\n");
	printf("\n");
	return 0;
}

int main(int argc, char *argv[])
{
	int n;

	for(n = 1; n < argc; n++) {
		char *cmd = argv[n], *arg = argv[n+1];
		while(*cmd == '-') cmd++;
		switch(*cmd) {
			case 'h': return help(argv[0], arg);
			case 'f':
				if (ifn != NULL) {
					fprintf(stderr, "Error at %s %s: only one -f is supported.\n", argv[n], argv[n+1]);
					return 1;
				}
				ifn = arg;
				break;
			case 'c':
				if (icmd != NULL) {
					fprintf(stderr, "Error at %s %s: only one -c is supported.\n", argv[n], argv[n+1]);
					return 1;
				}
				icmd = arg;
				break;
		}
	}

	if (ifn != NULL) {
		struct stat st;
		if (stat(ifn, &st) != 0)
			fprintf(stderr, "Warning: input file %s not found\n", ifn);
	}

	projview_init(&prj);

	/* better defaults for viewing a board */
	prj.ax = 0.7;
	prj.az = 1.7;
	prj.zoom = 60;

	check_file();
	main_loop();
	return 0;
}
